void pollard_phi(const int phase1, const double phase2)
{
  // Pollard-Phi Methode modifiziert in drei Phasen:
  // Phase 0: alle kleine Zahlen bis konstanter Obergrenze
  // Phase 1: (im wesentlichen) Primzahlen bis phase1 als Obergrenze
  //          (Primfaktor-1) kann/darf in alle diese Primzahlen zerfallen
  // Phase 2: (im wesentlichen) Primzahlen bis phase2 als Obergrenze
  //          bis auf einen(!) aus Phase 2 mssen alle anderen
  //          Faktoren von (Primfaktor-1) bereits vorher entdeckt worden
  //          sein, um Primfaktor entdecken zu knnen.
  //          Es stehen diverse Methoden fr Phase 2 zur Verfgung,
  //          die jedoch momentan nur ber eine nderung in diesem Sourcecode
  //          aktiviert werden knnen.
  //           - phi-phi continuation
  //           - standard continuation
  //           - improved standard continuation
  //           - improved standard continuation with pairing
  //           - improved standard continuation with pairing & fft

  mpz_t a,b;
  mpz_init(a); mpz_init(b);

  int i;
  short int d;
  
  // Phase 0: kleine Zahlen (mit Primzahlpotenzen)
  cout << "Phi-Phase 0" << endl;
  mpz_set_ui(a,2);
  for (i=5000; i<10000; i++)
    {
      mpz_powm_ui(a,a,i,n);
      mpz_sub_ui(x,a,1); mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)  /* Teiler gefunden? */
	{
	  cout << "Faktor gefunden in Phase 0: i=" << i << endl;
	  if (mpz_cmp(x,n)==0) // gefundener Faktor = n ?
	    {
	      cout << "Gefundene Faktorisierung ist trivial. Breche ab." << endl;
	      goto fertig; // trivialer Teiler -> Faktoren anders aufrollen oder Verfahren abbrechen
	    }
	  mpz_divexact(n,n,x);
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi0]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      if (mpz_probab_prime_p(n,probab_prime_checks))
               {
		 mpz_swap(n,x);
		 cout << x << " ist Faktor." << endl;
	         Factorization_to_file << MAL() << x << " [phi0/factorswap]" << flush;
               }
              else
               {
	         Factorization_to_file << MAL() << x << " [phi0] [composite]" << flush;
               }
	    }
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig;
	}
    }
  
#if 1 /* normale phi-Phase 1 */
  {
   // Phase 1: Primzahlen ohne Potenzen (aber knnen in beliebiger Kombination p-1 teilen)
   const int D=2*3*5*7*11*13;
   i=i-(i%D); // i mit Eigenschaft i=5 (mod 6) fr schnellere Primzahlermittlung
   cout << "Phi-Phase 1" << endl;
   while (i<=phase1)
    {
      // Primzahlen aussieben
      bool sieve[D];
      for (int j=1; j<=D; j+=2) sieve[j]=true; // Sieb initialisieren
      for (int p=5, dp=2; p*p<D+i; p+=dp, dp=6-dp)
       for (int j=p-(i%p); j<D; j+=p) sieve[j]=false;
       /* 
         Bemerkung:
          a) p*p<D+i ist als Primzahlbedingung fr das Intervall
             etwas zu streng, aber strt wegen Phase 0 nicht.
             (-> Primzahlen unter sqrt(D) werden nicht erkannt) 
          b) p-(i%p) ist zwar p, wenn p teilt i;
             aber: den Index 0 brauchen wir spter nicht!
       */
            
      for (int j=1, dj=4; j<D; j+=dj, dj=6-dj)
       {
         //if (sieve[j]!=is_prime(i+j))
         //  { cerr << "keine bereinstimmung: " << i+j << endl; }
         if (sieve[j]) my_mpz_powm_ui(a,i+j,n);
       }
      /* Bemerkung:
         Nicht durch die Literatur verwirren lassen:
         Das akkumulierte Produkt der (a-1)-Werte mu natrlich
         nicht gebildet werden. Das wre eine total berflssige
         Aktion, da ein Teiler auch bei nachfolgenden a's erhalten
         bleiben mu (aufgrund des kleinen Fermatschen Satzes!).
         Falls sich andererseits mehrere Teiler akkumulieren, so kann
         auch das akkumulierte Produkt dies nicht verhindern...
         Keine Ahnung, warum ich das akkumulierte Produkt frher hier
         gebildet habe, offenbar habe ich mich verwirren lassen...
         (Ist ja auch wie mit den Elefanten: "Warum klatschst du denn
          dauernd in die Hnde?" - "Um die Elefanten zu verscheuchen!" -
          "Welche Elefanten?" - "Siehtst du! Alle verscheucht!")
      */
      // Schleife vermeidet hufige gcd-Berechnung
      // in seltenen Fllen knnen aber mehrere Faktoren zusammenfallen
      // -> Phi-Methode versagt dann (ggf. Schleifendurchlufe vermindern)

      i+=D;
      cout << "Phase1: " << i << "/" << phase1 << "\r" << flush;

      mpz_sub_ui(x,a,1); mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)
        {	  
	  cout << "Faktor gefunden in Phase 1: i=" << i << endl;
	  if (mpz_cmp(x,n)==0) // gefundener Faktor = n ?
	    {
	      cout << "Gefundene Faktorisierung ist trivial. Breche ab." << endl;
	      goto fertig; // trivialer Teiler -> Faktoren anders aufrollen oder Verfahren abbrechen
            }
	  mpz_divexact(n,n,x);
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi1]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      if (mpz_probab_prime_p(n,probab_prime_checks))
               {
		 mpz_swap(n,x);
		 cout << x << " ist Faktor." << endl;
	         Factorization_to_file << MAL() << x << " [phi1/factorswap]" << flush;
               }
              else
               {
	         Factorization_to_file << MAL() << x << " [phi1] [composite]" << flush;
               }
	    }
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig; 
	}
    }
   i--; d=4; while (!is_prime(i)) { i-=d; d=6-d; } // i auf vorige Primzahl adjustieren
   cout << "Phi-Phase 1 bis " << i << endl;
   if (i>=phase2) goto fertig;
  }
#else /* modifizierte phi-Phase 1 */
  {
   // Phase 1: Primzahlen ohne Potenzen (aber knnen in beliebiger Kombination p-1 teilen)
   const unsigned int K=2;
   unsigned int k=0;
   mpz_t base[K];
   for (unsigned int j=0; j<K; ++j) mpz_init_set_ui(base[j],1);

   const int D=2*3*5*7*11*13;
   i=i-(i%D); // i mit Eigenschaft i=5 (mod 6) fr schnellere Primzahlermittlung
   cout << "Phi-Phase 1 (modified)" << endl;
   while (i<=phase1)
    {
      // Primzahlen aussieben
      bool sieve[D];
      for (int j=1; j<=D; j+=2) sieve[j]=true; // Sieb initialisieren
      for (int p=5, dp=2; p*p<D+i; p+=dp, dp=6-dp)
       for (int j=p-(i%p); j<D; j+=p) sieve[j]=false;
       /* 
         Bemerkung:
          a) p*p<D+i ist als Primzahlbedingung fr das Intervall
             etwas zu streng, aber strt wegen Phase 0 nicht.
             (-> Primzahlen unter sqrt(D) werden nicht erkannt) 
          b) p-(i%p) ist zwar p, wenn p teilt i;
             aber: den Index 0 brauchen wir spter nicht!
       */
      
      mpz_set_ui(b,1);
      for (int j=1, dj=4; j<D; j+=dj, dj=6-dj)
       {
         //if (sieve[j]!=is_prime(i+j))
         //  { cerr << "keine bereinstimmung: " << i+j << endl; }
         if (sieve[j]) 
          {
            my_mpz_powm_ui(a,i+j,n);
            mpz_sub(x,a,base[k]); mpz_mul(b,b,x); mpz_mod(b,b,n);
            mpz_set(base[k],a); k++; if (k>=K) k=0;
            //cout << "i+j=" << i+j << "b=" << b << endl;
          }
       }
      // Schleife vermeidet hufige gcd-Berechnung
      // in seltenen Fllen knnen aber mehrere Faktoren zusammenfallen
      // -> Phi-Methode versagt dann (ggf. Schleifendurchlufe vermindern)

      i+=D;
      cout << "Phase1: " << i << "/" << phase1 << "\r" << flush;

      mpz_gcd(x,b,n);
      if (mpz_cmp_ui(x,1)!=0)
        {	  
	  cout << "Faktor gefunden in Phase 1: i=" << i << endl;
	  if (mpz_cmp(x,n)==0) // gefundener Faktor = n ?
	    {
	      cout << "Gefundene Faktorisierung ist trivial. Breche ab." << endl;
	      goto fertig; // trivialer Teiler -> Faktoren anders aufrollen oder Verfahren abbrechen
            }
	  mpz_divexact(n,n,x);
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi1]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      if (mpz_probab_prime_p(n,probab_prime_checks))
               {
		 mpz_swap(n,x);
		 cout << x << " ist Faktor." << endl;
	         Factorization_to_file << MAL() << x << " [phi1/factorswap]" << flush;
               }
              else
               {
	         Factorization_to_file << MAL() << x << " [phi1] [composite]" << flush;
               }
	    }
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig; 
	}
    }
   i--; d=4; while (!is_prime(i)) { i-=d; d=6-d; } // i auf vorige Primzahl adjustieren
   cout << "Phi-Phase 1 bis " << i << endl;
   for (unsigned int j=0; j<K; ++j) mpz_clear(base[j]);
   if (i>=phase2) goto fertig;
  }
#endif

#if 0 /* standard continuation */
  {
   // Phase 2: isolierte (Prim-)zahl: p-1 mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Phi-Phase 2" << endl;
   mpz_set(b,a); // neue Basis fr phi-Methode

   // Precomputing most common used exponents:
   const int pre_exp_size = 100;
   mpz_t bpow[pre_exp_size];
   for (int j=2; j<pre_exp_size; j+=2)
    {
      mpz_init(bpow[j]);
      mpz_powm_ui(bpow[j],b,j,n);
    }  

   while (!probab_prime(i)) { i+=d; d=6-d; }
   mpz_powm_ui(a,a,i,n); mpz_sub_ui(x,a,1);
  
   while (i<=phase2) 
    { 
      int runden=5000;
      while (runden>0)
	{
	  int i_prev=i;
          do { i+=d; d=6-d; } while (i%5==0 || i%7==0 || i%11==0 || i%13==0 ||
                                     i%17==0 || i%19==0 || i%23==0 || i%29==0 ||
	           		     i%31==0 || i%37==0 || i%41==0 || i%43==0 ||
	           		     i%47==0 || i%53==0 || i%59==0 || i%61==0);
				     // || powmod(2,i-1,i)!=1);
	 // obige do-Schleife soll mglichst schnell viele zusammengesetzte
         // Zahlen ausschlieen, (Fermat-Test hier zu teuer!) 
          if (i-i_prev<pre_exp_size) mpz_mul(a,a,bpow[i-i_prev]);
          else
           { // not a precomputed power of b:
	     cerr << "not a precomputed exponent " << i-i_prev << endl;
             mpz_powm_ui(y,b,i-i_prev,n);
	     mpz_mul(a,a,y);
           }
          mpz_mod(a,a,n);
	  mpz_sub_ui(y,a,1); mpz_mul(x,x,y); mpz_mod(x,x,n);
	  --runden;
	}
      cout << "Phase2: " << i << "/" << phase2 << "\r" << flush;
      mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)
	{
	  cout << "Faktor gefunden in Phase 2: i=" << i << endl;
	  mpz_divexact(n,n,x);
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      if (mpz_probab_prime_p(n,probab_prime_checks))
               {
		 mpz_swap(n,x);
		 cout << x << " ist Faktor." << endl;
	         Factorization_to_file << MAL() << x << " [phi2/factorswap]" << flush;
               }
              else
               {
	         Factorization_to_file << MAL() << x << " [phi2] [composite]" << flush;
               }
	    }
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig_phase2; 
	}
    }

 fertig_phase2:
  cout << "Phi-Phase 2 bis " << i << endl;
  for (int j=2; j<pre_exp_size; j+=2)
    mpz_clear(bpow[j]); // freeing memory for precomputed exponents
  }
#endif

#if 0 /* improved standard continuation */
  {
   // Phase 2: isolierte (Prim-)zahl: p-1 mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Phi-Phase 2 (improved)" << endl;
   mpz_set(b,a); // neue Basis fr phi-Methode

   // Precomputing most common used exponents:
   const int pre_exp_size = 2*3*5*7*11*13;
   mpz_t bpow[pre_exp_size];
   for (int j=1; j<pre_exp_size; j+=2)
    {
      if ((j%3)&&(j%5)&&(j%7)&&(j%11)&&(j%13))
       {
         mpz_init(bpow[j]);
         mpz_powm_ui(bpow[j],b,j-1,n);
       }
    }

   mpz_powm_ui(b,b,pre_exp_size,n); // Intervall-Multiplikator  b := b^pre_exp_size mod n;

   i=((i/pre_exp_size)*pre_exp_size)-1; d=2;
   mpz_powm_ui(a,a,i,n); mpz_sub_ui(x,a,1);

   int next_i = i+1+pre_exp_size;
   mpz_mul(a,a,b); mpz_mod(a,a,n); // ein Intervall weiterspringen

   bool sieve[pre_exp_size];
   for (int i1=pre_exp_size-1; i1>=0; i1-=2) sieve[i1]=false;
   for (int p=5, dp=2; p<pre_exp_size && p*p<next_i; p+=dp, dp=6-dp )
    for (int i1=p-((next_i-pre_exp_size)%p); i1<pre_exp_size; i1+=p) sieve[i1]=true;
   
   while (i<=phase2) 
    { 
      int runden=10000;
      while (runden>0)
	{
          // nchste Primzahl ermitteln...
          do { i+=d; d=6-d; } while (i<next_i && sieve[i%pre_exp_size]);
	  // obige do-Schleife soll mglichst schnell viele zusammengesetzte
	  // Zahlen ausschlieen (mittels Siebverfahren)
          // (ein Fermat-Test wre hier zu teuer!)

	  while (i>next_i) // "if" tut's bei gengend groen pre_exp_size auch...
	    {
              // ein Intervall weiterspringen
              next_i+=pre_exp_size;
              mpz_mul(a,a,b); mpz_mod(a,a,n);
              for (int i1=pre_exp_size-1; i1>=0; i1-=2) sieve[i1]=false;
              for (int p=5, dp=2; p<pre_exp_size && p*p<next_i; p+=dp, dp=6-dp )
               for (int i1=p-((next_i-pre_exp_size)%p); i1<pre_exp_size; i1+=p) sieve[i1]=true;
	    }
          mpz_sub(y,a,bpow[next_i-i]);
          mpz_mul(x,x,y); mpz_mod(x,x,n);
	  --runden;
	}
      cout << "Phase2: " << i << "/" << phase2 << "\r" << flush;
      mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)
	{
	  cout << "Faktor gefunden in Phase 2: i=" << i << endl;
	  mpz_divexact(n,n,x);
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      if (mpz_probab_prime_p(n,probab_prime_checks))
               {
		 mpz_swap(n,x);
		 cout << x << " ist Faktor." << endl;
	         Factorization_to_file << MAL() << x << " [phi2/factorswap]" << flush;
               }
              else
               {
	         Factorization_to_file << MAL() << x << " [phi2] [composite]" << flush;
               }
	    }
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig_phase2;

	  // nun noch die Exponenten anpassen (damit's schneller geht)
	  for (int j=1; j<pre_exp_size; j+=2)
           if ((j%3)&&(j%5)&&(j%7)&&(j%11)&&(j%13)) mpz_mod(bpow[j],bpow[j],n);
	}
    }

 fertig_phase2:
  cout << "Phi-Phase 2 bis " << i << endl;
  for (int j=1; j<pre_exp_size; j+=2)
   if ((j%3)&&(j%5)&&(j%7)&&(j%11)&&(j%13))
    mpz_clear(bpow[j]); // freeing memory for precomputed exponents
  }
#endif

#if 0 /* improved standard continuation with pairing */
  {
   // Phase 2: isolierte (Prim-)zahl: p-1 mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Phi-Phase 2 (improved with pairing)" << endl;

   // in a steht die Start-Basis fr Phase 2

  /* Grundidee:
     Ausnutzung der folgenden Gleichheit:
     (a^(i+d) -1) * (a^(i-d) -1) = a^i * ( (a^i +a^(-i)) - (a^d +a^(-d)) )
     
     1. Der Multiplikatior a^i kann weggelassen werden, da er das
        Teilerverhalten beim ggT nicht beeinflut.
     2. In Intervallen der Gre D wird nun i erhht.
         -> Vorberechnung von (a^d +a^(-d)) fr 0<d<D,
         -> auerdem: pro Intervall eine Bestimmung von (a^i + a^(-i))
     3. a^(i+D) lt sich sukzessiv ber a^(i+D)=a^i * a^D bestimmen;
        a^(-(i+D)) lt sich wegen a^(-i)=(a^(-1))^i analog bestimmen.
        -> pro Intervall zwei Multiplikationen, um a^(i+D) + a^(-(i+D))
           zu bestimmen.

    Bemerkung I:
      Die hierauf aufbauende Paarbildungsmethode konstruiert nicht
      die durch andere Methoden erreichbaren Primzahlpaare, sie ist
      aber dafr vergleichsweise einfach zu implementieren und sie
      kommt mit weniger Speicherplatz als die improved-standard-continuation
      aus. Sobald mehrere Primzahlpaare innerhalb eines Intervalls
      auftauchen, ist sie schneller als die improved standard continuation.
   
    Bemerkung II: (zu 3.)
      a^i+a^(-i) lt sich -- wegen x^2-x(a^1+a^(-1))+1=(x-a)*(x-a^(-1)) --
      auch ber Lucasfolgen effizient berechnen:
      a^i+a^(-i)=Vi(a^1+a^(-1))=:V[i];
      a^(i+D)+a^(-(i+D)))=V[i+D]=V[i]*V[D]-V[i-D].
      Man kann also die Anzahl der erforderlichen Multiplikationen pro
      Intervallsprung von 2 auf 1 reduzieren.
      Damit wren wir dann (mit anderer Herleitung und mit leichten
      Modifikationen) wieder auf bereits bekannte Fortsetzungen gestoen
      (vgl. Peter L. Montgomery, "Speeding the Pollard and Elliptic Curve
       Methods of Factorization", MathComp Vol 48 (177), Jan 1987, p.243-264;
       insbesondere Abschnitte 4, 5 und 9).
  */

   mpz_t b_inv, d, d_inv, b_plus_b_inv, d_plus_d_inv, prev_b_plus_b_inv;
   mpz_init(b_inv); mpz_init(d); mpz_init(d_inv); mpz_init(b_plus_b_inv);
   mpz_init(d_plus_d_inv); mpz_init(prev_b_plus_b_inv);

   // im folgenden werden b,b_inv, d,d_inv nicht mehr aktualisiert,
   // da die zugehrigen Summen direkt durch Lucasfolgen berechnet werden.

   // Precomputing most common used values:
   const int D = 2*3*5*7*11*13; // 2*3 unbedingt notwendig!
   mpz_t valtab[D/2];

   mpz_set(b,a); mpz_invert(b_inv,b,n); mpz_mod(b_inv,b_inv,n);
   mpz_powm_ui(d,b,2,n); mpz_powm_ui(d_inv,b_inv,2,n);
   mpz_add(d_plus_d_inv,d,d_inv); mpz_mod(d_plus_d_inv,d_plus_d_inv,n);
   mpz_add(b_plus_b_inv,b,b_inv); mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
   mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
   for (int j=1; j<D/2; j+=2)
    {
      if (gcd(j,D)==1) mpz_init_set(valtab[j],b_plus_b_inv);
#if 0
       {
         mpz_init(valtab[j]);
         //mpz_powm_ui(x,a,j,n);
         //mpz_invert(y,x,n);
         //mpz_add(x,x,y);
         mpz_add(x,b,b_inv);
         mpz_mod(valtab[j],x,n); // valtab[j] = a^j+a^(-j) mod n

         cout << valtab[j] << " = " << b_plus_b_inv << endl;
         cout << endl;
         char ch;
         cin >> ch;
       }

      mpz_mul(b,b,d); mpz_mod(b,b,n);
      mpz_mul(b_inv,b_inv,d_inv); mpz_mod(b_inv,b_inv,n);
#endif
      mpz_set(x,prev_b_plus_b_inv); mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
      mpz_mul(b_plus_b_inv,b_plus_b_inv,d_plus_d_inv);
      mpz_sub(b_plus_b_inv,b_plus_b_inv,x);
      mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
    }

   i=i-(i%D); // Startpunkt ermitteln
   mpz_powm_ui(d,a,D,n); mpz_invert(d_inv,d,n); mpz_mod(d_inv,d_inv,n);
   mpz_add(d_plus_d_inv,d,d_inv); mpz_mod(d_plus_d_inv,d_plus_d_inv,n);
   mpz_powm_ui(b,a,i,n); mpz_invert(b_inv,b,n); mpz_mod(b_inv,b_inv,n);
   mpz_add(b_plus_b_inv,b,b_inv); mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
   mpz_mul(x,b,d_inv); mpz_mod(x,x,n);
   mpz_mul(y,b_inv,d); mpz_mod(y,y,n);
   mpz_add(prev_b_plus_b_inv,x,y); mpz_mod(prev_b_plus_b_inv,prev_b_plus_b_inv,n);

   i-=D/2; // Intervall von [i-D/2,i+D/2] auf [i,i+D] normieren
   while (i<=phase2)
    {
      int runden = 10000; // Anzahl Runden bis zu gcd und Statusausgabe
      mpz_set_ui(y,1); // Produktsammler fr gcd

      while (i<=phase2 && runden>0)
       {
         // jetzt das Intervall [i,i+D] abarbeiten
#if 0
         mpz_add(x,b,b_inv); mpz_mod(x,x,n);
         cout << x << endl;
         cout << b_plus_b_inv << endl;
         cout << endl;
         char ch; cin >> ch;
#endif
         // Primzahlen aussieben
         bool sieve[D];
         for (int j=1-(i%2); j<=D; j+=2) sieve[j]=true; // Sieb initialisieren
         for (int p=5, dp=2; p*p<D+i; p+=dp, dp=6-dp)
           for (int j=p-(i%p); j<D; j+=p) sieve[j]=false;
           /* 
             Bemerkung:
              p-(i%p) ist zwar p, wenn p teilt i;
              aber: den Index 0 brauchen wir spter nicht!
           */

         // Zahlenpaare i+D/2-j,i+D/2+j mit mindestens einer Primzahl auslesen
         for (int j=1, dj=4; j<D/2; j+=dj, dj=6-dj)
         if (sieve[D/2-j] || sieve[D/2+j])
          {
#if 0
            if (!( is_prime(i+D/2-j) || is_prime(i+D/2+j) ))
             {
               cout << "FEHLER!!"
                    << " i= " << i 
	  	    << " j= " << j 
		    << " D/2= " << D/2 
                    << endl;
               exit(1);
             }
#endif
            mpz_sub(x,b_plus_b_inv,valtab[j]); // valtab ohne Displacement!
            mpz_mul(y,y,x); mpz_mod(y,y,n);
            runden--;
          }

         // Werte fr nchstes Intervall bestimmen...
#if 0
         mpz_mul(b,b,d); mpz_mod(b,b,n);
         mpz_mul(b_inv,b_inv,d_inv); mpz_mod(b_inv,b_inv,n);
#endif
         i+=D; // neues Intervall (bis i haben wir nun abgearbeitet)
         mpz_set(x,prev_b_plus_b_inv); mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
         mpz_mul(b_plus_b_inv,b_plus_b_inv,d_plus_d_inv);
         mpz_sub(b_plus_b_inv,b_plus_b_inv,x);
         mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
       }

      cout << "Phase2: " << i << "/" << phase2 << "\r" << flush;

      // nun (erst) den ggT ermitteln
      mpz_gcd(x,y,n);
      if (mpz_cmp_ui(x,1)!=0)
	{
	  cout << "Faktor gefunden in Phase 2: i=" << i << endl;
	  mpz_divexact(n,n,x);
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      if (mpz_probab_prime_p(n,probab_prime_checks))
               {
		 mpz_swap(n,x);
		 cout << x << " ist Faktor." << endl;
	         Factorization_to_file << MAL() << x << " [phi2/factorswap]" << flush;
               }
              else
               {
	         Factorization_to_file << MAL() << x << " [phi2] [composite]" << flush;
               }
	    }
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig_phase2;

	  // nun noch die Exponenten anpassen (damit's schneller geht)
	  for (int j=1; j<D/2; j+=2)
           if (gcd(j,D)==1) mpz_mod(valtab[j],valtab[j],n);
	}

    }

 fertig_phase2:
  cout << "Phi-Phase 2 bis " << i << endl;
  mpz_clear(b_inv); mpz_clear(d); mpz_clear(d_inv); mpz_clear(b_plus_b_inv);
  mpz_clear(d_plus_d_inv); mpz_clear(prev_b_plus_b_inv);
  for (int j=1; j<D/2; j+=2)
   if (gcd(j,D)==1) 
    mpz_clear(valtab[j]); // freeing memory for precomputed values
  }
#endif

#if 1 /* improved standard continuation with pairing & fft */
  {
   double ii=i;
   //const double i = -1; // uncomment this just to be sure that i isn't used anymore...

   // Phase 2: isolierte (Prim-)zahl: p-1 mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Phi-Phase 2 (improved with pairing & fft)" << endl;

   /* zur Bedeutung nachfolgender Werte siehe elliptische Kurven,
      fft-continuation */
   unsigned int pg=256, pms=2*1050;
   get_fft_parameter(phase2,pg,pms); // Parameter (pg,pms) fr "phase2" holen

   const int D = pms;
   const int Polynom_Index = pg;

   cout << "Polynomgre=" << Polynom_Index << endl;
   cout << "Einzel-Intervallsize=" << D << endl;

   mpz_t sammel[Polynom_Index];
   for (int j=0; j<Polynom_Index; ++j) mpz_init(sammel[j]);

   // in a steht die Start-Basis fr Phase 2

  /* Grundidee:
     Ausnutzung der folgenden Gleichheit:
     (a^(i+d) -1) * (a^(i-d) -1) = a^i * ( (a^i +a^(-i)) - (a^d +a^(-d)) )
     
     1. Der Multiplikatior a^i kann weggelassen werden, da er das
        Teilerverhalten beim ggT nicht beeinflut.
     2. In Intervallen der Gre D wird nun i erhht.
         -> Vorberechnung von (a^d +a^(-d)) fr 0<d<D,
         -> auerdem: pro Intervall eine Bestimmung von (a^i + a^(-i))
     3. a^(i+D) lt sich sukzessiv ber a^(i+D)=a^i * a^D bestimmen;
        a^(-(i+D)) lt sich wegen a^(-i)=(a^(-1))^i analog bestimmen.
        -> pro Intervall zwei Multiplikationen, um a^(i+D) + a^(-(i+D))
           zu bestimmen.

    Bemerkung I:
      Die hierauf aufbauende Paarbildungsmethode konstruiert nicht
      die durch andere Methoden erreichbaren Primzahlpaare, sie ist
      aber dafr vergleichsweise einfach zu implementieren und sie
      kommt mit weniger Speicherplatz als die improved-standard-continuation
      aus. Sobald mehrere Primzahlpaare innerhalb eines Intervalls
      auftauchen, ist sie schneller als die improved standard continuation.
   
    Bemerkung II: (zu 3.)
      a^i+a^(-i) lt sich -- wegen x^2-x(a^1+a^(-1))+1=(x-a)*(x-a^(-1)) --
      auch ber Lucasfolgen effizient berechnen:
      a^i+a^(-i)=Vi(a^1+a^(-1))=:V[i];
      a^(i+D)+a^(-(i+D)))=V[i+D]=V[i]*V[D]-V[i-D].
      Man kann also die Anzahl der erforderlichen Multiplikationen pro
      Intervallsprung von 2 auf 1 reduzieren.
      Damit wren wir dann (mit anderer Herleitung und mit leichten
      Modifikationen) wieder auf bereits bekannte Fortsetzungen gestoen
      (vgl. Peter L. Montgomery, "Speeding the Pollard and Elliptic Curve
       Methods of Factorization", MathComp Vol 48 (177), Jan 1987, p.243-264;
       insbesondere Abschnitte 4, 5 und 9).
  */

   mpz_t b_inv, d, d_inv, b_plus_b_inv, d_plus_d_inv, prev_b_plus_b_inv;
   mpz_init(b_inv); mpz_init(d); mpz_init(d_inv); mpz_init(b_plus_b_inv);
   mpz_init(d_plus_d_inv); mpz_init(prev_b_plus_b_inv);

   // im folgenden werden b,b_inv, d,d_inv nicht mehr aktualisiert,
   // da die zugehrigen Summen direkt durch Lucasfolgen berechnet werden.

   // Precomputing most common used values:
   int Polynom_Index_i = 0;
   mpz_set(b,a); mpz_invert(b_inv,b,n); mpz_mod(b_inv,b_inv,n);
   mpz_powm_ui(d,b,2,n); mpz_powm_ui(d_inv,b_inv,2,n);
   mpz_add(d_plus_d_inv,d,d_inv); mpz_mod(d_plus_d_inv,d_plus_d_inv,n);
   mpz_add(b_plus_b_inv,b,b_inv); mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
   mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
   for (int j=1; j<D/2; j+=2)
    {
      if (gcd(j,D)==1) mpz_set(sammel[Polynom_Index_i++],b_plus_b_inv);

      mpz_set(x,prev_b_plus_b_inv); mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
      mpz_mul(b_plus_b_inv,b_plus_b_inv,d_plus_d_inv);
      mpz_sub(b_plus_b_inv,b_plus_b_inv,x);
      mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
    }

   cout << "unvergrerte Polynomgre: " << Polynom_Index_i << endl;
   polynomial::TPolynom Polynom = NULL; // hier steht das anschlieend konstruierte Polynom drin
   // Rest bis zur nchsten Zweierpotenz mit Nullen auffllen
   /* Bemerkung: Besser wre es, Punkte auerhalb des Suchbereichs
                 zu whlen (Motto: doppeltes, aber auerhalb des
                 Suchbereichs lchriges Netz)
    */
   { 
     int anz_Punkte=Polynom_Index_i;    
     while (anz_Punkte<Polynom_Index) mpz_set_ui(sammel[anz_Punkte++],0);
     Polynom_Index_i=polynomial::Polynom_aus_Nullstellen_konstruieren(Polynom,sammel,anz_Punkte,n);
     cout << "Polynom erzeugt. Polynomindex = " << Polynom_Index_i << endl;
   }

   ii=floor(ii/D)*D; // i=i-(i%D); // Startpunkt ermitteln
   mpz_powm_ui(d,a,D,n); mpz_invert(d_inv,d,n); mpz_mod(d_inv,d_inv,n);
   mpz_add(d_plus_d_inv,d,d_inv); mpz_mod(d_plus_d_inv,d_plus_d_inv,n);
   mpz_set_d(x,ii); mpz_powm(b,a,x,n);
   mpz_invert(b_inv,b,n); mpz_mod(b_inv,b_inv,n);
   mpz_add(b_plus_b_inv,b,b_inv); mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
   mpz_mul(x,b,d_inv); mpz_mod(x,x,n);
   mpz_mul(y,b_inv,d); mpz_mod(y,y,n);
   mpz_add(prev_b_plus_b_inv,x,y); mpz_mod(prev_b_plus_b_inv,prev_b_plus_b_inv,n);

   ii-=D/2; // Intervall von [i-D/2,i+D/2] auf [i,i+D] normieren
   while (ii<=phase2)
    {
      cout << "Phase2: " << ii << "/" << phase2 << endl << flush;
      cout << "Sammle Werte..." << endl;

      //Werte sammeln
      int sammel_index=0;
      while (sammel_index<(Polynom_Index_i-1)/2)
       {
         mpz_set(sammel[sammel_index++],b_plus_b_inv);
         // Werte fr nchstes Intervall bestimmen...
         ii+=D; // neues Intervall (bis i haben wir nun abgearbeitet)
         mpz_set(x,prev_b_plus_b_inv); mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
         mpz_mul(b_plus_b_inv,b_plus_b_inv,d_plus_d_inv);
         mpz_sub(b_plus_b_inv,b_plus_b_inv,x);
         mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
       }

      // Polynomauswertung starten
      cout << "Phase2: " << ii << "/" << phase2 << endl << flush;
      cout << "starte Polynomauswertung" << endl;
      cout << "Parameter: Polynomindex: " << Polynom_Index_i
           << ", Anzahl der Punkte: " << sammel_index << endl;
      polynomial::multipoint_eval(sammel,Polynom,Polynom_Index_i,sammel,sammel_index,n);
      cout << "Polynomauswertung beendet. Teste auf ggT." << endl;
      mpz_set(y,sammel[0]);
      for (int j=1; j<sammel_index; ++j)
        { mpz_mul(y,y,sammel[j]); mpz_mod(y,y,n); }
      // nun (erst) den ggT ermitteln
      mpz_gcd(x,y,n);
      if (mpz_cmp_ui(x,1)==0) continue;
      
      // nun ist/sind ein/mehrere Teiler(in/nen) gefunden,
      // mgliches Problem: vielleicht mehrere Teiler akkumuliert.
      // Da es aber nur log-viele Teiler geben kann, ist dieser
      // Abschnitt nicht zeitkritisch; also strt es wenig, alle
      // Werte nochmal durchzugehen (dies senkt die Wahrscheinlichkeit,
      // akkumulierte Teiler zu erhalten).
    
      for (int j=1; j<sammel_index; ++j)
       {
	 mpz_gcd(x,sammel[j],n);
         if (mpz_cmp_ui(x,1)!=0)
	  {
	   cout << "Faktor gefunden in Phase 2: i=" << ii << endl;
	   mpz_divexact(n,n,x);
	   if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2]" << flush;
	    }
	   else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      if (mpz_probab_prime_p(n,probab_prime_checks))
               {
		 mpz_swap(n,x);
		 cout << x << " ist Faktor." << endl;
	         Factorization_to_file << MAL() << x << " [phi2/factorswap]" << flush;
               }
              else
               {
	         Factorization_to_file << MAL() << x << " [phi2] [composite]" << flush;
               }
	    }
	   if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig_phase2;
           // anschlieend noch Polynom vereinfachen:
	   for (int j=0; j<Polynom_Index_i; ++j) mpz_mod(Polynom[j],Polynom[j],n);
	  }
       }

    }

 fertig_phase2:
  cout << "Phi-Phase 2 bis " << ii << endl;
  mpz_clear(b_inv); mpz_clear(d); mpz_clear(d_inv); mpz_clear(b_plus_b_inv);
  mpz_clear(d_plus_d_inv); mpz_clear(prev_b_plus_b_inv);
  for (int j=0; j<Polynom_Index; ++j) mpz_clear(sammel[j]);
  for (int j=0; j<Polynom_Index_i; ++j) mpz_clear(Polynom[j]);
  delete [] Polynom;
  }
#endif


 fertig:
  mpz_clear(a); mpz_clear(b);
}
