#ifndef qsieve_header
#define qsieve_header


inline double sqrt(const int i)
{
  return sqrt(static_cast<double>(i));
}

inline double sqrt(const unsigned int i)
{
  return sqrt(static_cast<double>(i));
}

inline double sqrt(const long long i)
{
  return sqrt(static_cast<double>(i));
}

inline double log(const int i)
{
  return log(static_cast<double>(i));
}

inline double log(const long long i)
{
  return log(static_cast<double>(i));
}

#include "TinyVector.H"
typedef CTinyVector<FBsizetype> CTinyFBsizetypeVector;

const int no_factor = -1;
const int dynamic_factor = -2;
const int cached_dynamic_factor = -3;
const int special_factor = -4;

class CGleichungszeile
/* CGleichungszeile verwaltet eine Relation der Form
       Delta^2 = Produkt von Faktoren (mod N)
   Ziel ist es, die Menge der Faktoren durch Verknpfung von Relationen
   zu leeren. Das Produkt wird dann 1.
   Delta pat sich den Vernderungen dynamisch an.
*/ 
{
 public:
  int relevanter_Faktor;
 private:
  CTinyFBsizetypeVector *Relation_sparse;
  myBitString *Relation_dense;
  mpz_t Delta; // dynamischer Delta-Wert
 public:
  inline CGleichungszeile() /* Konstruktor */
    {
      relevanter_Faktor=no_factor;
      Relation_sparse = new CTinyFBsizetypeVector;
      Relation_dense = NULL;
      mpz_init_set_ui(Delta,1);
    }
 public:
  inline ~CGleichungszeile() /* Destruktor */
    {
      if (Relation_sparse) { delete Relation_sparse; Relation_sparse=NULL; }
      if (Relation_dense) { delete Relation_dense; Relation_dense=NULL; }
      mpz_clear(Delta);
      relevanter_Faktor=no_factor;
    }
  
  CGleichungszeile(const signed int Siebpos, const short int HitCount=0); /* Konstruktur */
  /* "Deltas" mit dem Wert Delta initialisieren,
     "Relationen" mit den Primfaktoren des reduzierten Quadrates fllen,
     deren Exponenten ungerade sind. */

 public:
  int groesster_Faktor_in_Relation() const
    /* (Index) des grten Faktors in der Relation */
    {
      if (Relation_sparse)
	if (Relation_sparse->empty()) return no_factor;
        else 
	  {
	    return Relation_sparse->last();
	  }
      else 
	{
	  return Relation_dense->last(1);
	  /*
	    int h = Relation_dense->last(1);
	    if (h>=0) return h;
	    else return no_factor;
	  */
	}
    }

  int zweitgroesster_Faktor_in_Relation() const
    /* (Index) des zweitgrten Faktors in der Relation */
    {
      if (Relation_sparse)
        if (Relation_sparse->size()>1)
          return Relation_sparse->v[Relation_sparse->size()-2];
        else
          return no_factor;
      else // Relation ist dense
        {
          signed int vorletzter_Faktor=Relation_dense->last(1);
          if (vorletzter_Faktor>=0)
           return Relation_dense->prev(vorletzter_Faktor,1);
          else
           return no_factor;
        }
    }
  

  inline unsigned int SizeOfRelation() const
    {
      if (Relation_sparse) return Relation_sparse->size();
      else return Relation_dense->count(1);
    }

  inline void optisize(void)
    {
      // Speicherbedarf der Relation optimieren (=minimieren)
      // 1. Falls sinnvoll, eine implizite Konvertierung dense<->sparse,
      // 2. Speicherbedarf innerhalb der Darstellung minimieren

      int size=SizeOfRelation();
      int largest=groesster_Faktor_in_Relation();
      int quotient = size ? largest/size : 0;

      if (Relation_sparse)
       {
         if (quotient<16-2) convert_Relation_to_dense();
         else Relation_sparse->optisize();
       }
      else // (Relation_dense)
       {
         if (quotient>16+2) convert_Relation_to_sparse();
         else Relation_dense->optisize();
       }
    }
  
  unsigned int SizeOfRelation_verknuepft_mit(const CGleichungszeile& GL2) const;
  
  void Ausgabe() const; /* Gleichungszeile ausgeben */
  void convert_Relation_to_dense(); /* mache "dense" aus "sparse" */
  void convert_Relation_to_sparse(); /* mache "sparse" aus "dense" */
  inline bool Relation_ist_leer() /* gibt zurck, ob die Relation leer ist */
    const
    {
      return (relevanter_Faktor==no_factor);
      // if (Relation_sparse) return Relation_sparse->empty();
      // else return Relation_dense->last(1)==-1; 
    }	
  void kombinieren(const CGleichungszeile& GL2);
  /* die Vereinigungsmengen mit GL2 bilden und daraus die
     Schnittmengen entfernen. (-> also: symmetrische Differenz bilden)
     gibt zurck, ob die Relationsmenge leer ist, d.h. eine Kombination
     quadratischer Reste mit dem gleichen Wert gefunden wurde. */

  // Routinen, die ein mehrfaches Kombinieren beschleunigen:
 protected:
  static unsigned int Exponentenfeld[MAX_Primbasis_size]; // (Startwert mu {0} sein!)
  static unsigned int multi_kombinieren_Zaehler; // (Startwert mu 0 sein!)
 public:
  void multi_kombinieren_init();
  void multi_kombinieren_main(const CGleichungszeile& GL2);
  void multi_kombinieren_exit();
  
  void Quadratische_Kongruenz_ausrechnen() const;
  bool check_Relation(const Tfactor Restfaktor) const;
  void Kongruenzcheck() const; // test, ob Kongruenz noch in Ordnung 
 private:
  inline void swap(CGleichungszeile &GL2)
    { /* Inhalte der beiden Gleichungszeilen (referenzsemantisch) tauschen */
      std::swap(Relation_sparse,GL2.Relation_sparse);
      std::swap(Relation_dense,GL2.Relation_dense);
      std::swap(relevanter_Faktor,GL2.relevanter_Faktor);
      mpz_swap(Delta,GL2.Delta);
    }
  CGleichungszeile operator= (const CGleichungszeile &GL)
    {
      cerr << "Zuweisung!!!! (nicht erlaubt)" << endl;
      exit(1);
      return GL;
    }
 public:
  inline const bool operator< (const CGleichungszeile &GL2) const
    {	
      return relevanter_Faktor < GL2.relevanter_Faktor;
    }
  streampos speichern(ostream &out, const Tfactor factor, const short int HitCount=0) const;
  inline streampos speichern(ostream &out, const int i=1, const short int HitCount=0) const
   {
     Tfactor DLP;
     DLP=i;
     return speichern(out,DLP,HitCount);
   };
  Tfactor kombinieren(istream &in, const streampos pos);
  Tfactor multi_kombinieren_main(istream &in, const streampos pos);
  Tfactor kombinieren(istream &in);
  Tfactor multi_kombinieren_main(istream &in);
  friend ostream& operator<< (ostream& ostr, const CGleichungszeile& GL);
  friend void Relation_einfuegen(CGleichungszeile *GL, const bool do_multi_kombinieren_init=true);
  friend void Special_Factoren_durch_Primfaktor_splitten(const int Primfaktor);
  friend void Special_Factoren_durch_Primfaktor_splitten_LAZY(const int Primfaktor);
  friend bool SpecialFactorRelation_vermerken(const Tfactor &DLP, CGleichungszeile *GL);
  friend bool SpecialFactorRelation_LAZY_vermerken(const Tfactor &DLP, const string &GL_String);
  friend void SpecialFactors_Zyklensuche();
  friend void bearbeite_daten_stream(iostream &daten);
  void Relation_optimieren(int bitpos=1);
};


ostream& operator<< (ostream& ostr, const CGleichungszeile& GL);
// Stream-Ausgabefunktion fr Gleichungszeile

//void Relation_einfuegen(CGleichungszeile *GL, const bool do_multi_kombinieren_init=true);
// fgt Gleichungszeile in das Gleichungssystem ein

void Special_Factoren_durch_Primfaktor_splitten(const int Primfaktor);
// Specialfaktoren liegen auerhalb der Primzahlbasis.
// Mit ihnen wird nicht gesiebt. Sie sind auerdem zusammengesetzt.
// Wenn es gelingt, von einem Specialfaktor einen Primfaktor aus der dynamischen FB
// abzusplitten, dann bestehen gute Chancen, dadurch einen neuen dynamischen Faktor
// zu erhalten oder gar einen Treffer im Gleichungssystem, wenn der entstandene
// dynamische Faktor bereits in der dynamischen Faktorbasis bekannt ist.


typedef list<CGleichungszeile> TZeilenliste;


class CPolynom
{
  /* Implementierung der Polynomberechnungen.
     Variablennotation hlt sich (einigermaen) an die Beschreibung
     aus 
     "The Multiple Polynomial Quadratic Sieve",
     Robert D. Silverman,
     Mathematics Of Computation, Vol. 48, January 1987, p. 329-339
  */  
private:
  int M;
  mpz_t A,B,C,D,h0,h1,h2;
  mpz_t kN, kN_div2, D2_inv_mod_kN;
  mpz_t A2;
public:
  inline CPolynom() // Konstruktur
    {
      mpz_init(A); mpz_init(B); mpz_init(C); mpz_init(D);
      mpz_init(h0); mpz_init(h1); mpz_init(h2);
      mpz_init(kN); mpz_init(kN_div2); mpz_init(D2_inv_mod_kN);
      mpz_init(A2);
    }
  inline ~CPolynom() // Destruktor
    {
      mpz_clear(A); mpz_clear(B); mpz_clear(C); mpz_clear(D);
      mpz_clear(h0); mpz_clear(h1); mpz_clear(h2);
      mpz_clear(kN); mpz_clear(kN_div2); mpz_clear(D2_inv_mod_kN);
      mpz_clear(A2);
    }
  void erstes_Polynom_ermitteln(const mpz_t &fuer_kN, const int M);
  void naechstes_Polynom_ermitteln(const int step=0);
  void Werte_holen(const int Siebpos, mpz_t &Wurzelwert, mpz_t &Q) const;
  void testen(const signed int Siebpos=0);
  void Daten_sichern(ostream &ostr); // fr Recovery etc.
  void Daten_laden(istream &in); // fr Recovery etc.
  inline bool operator < (const mpz_t &K) const 
    {
      return mpz_cmp(D, K)<0;
    }
  friend void Deltas_berechnen_fuer(const TDynamicFactorRelation &DynRel, const int offset);
  friend void alle_Deltas_neu_berechnen();
};


void Deltas_berechnen_fuer(const TDynamicFactorRelation &DynRel, const int offset);
void alle_Deltas_neu_berechnen();
  
#endif
