// compute number of partitions of n
// written by Thorsten Reinecke, 06.12.1999
// last change: 07.12.1999

/* 
  p(n) = number of unodered partitions of n
  = #solutions of x[1]+x[2]+...+x[n]=n, x[i]>=x[i+1], x[n+1]=0, i=1..n

  Reference:
    J.H. van Lint, R.M. Wilson: "A Course in Combinatorics",
    Cambridge University Press, 1992, p.132ff
*/
   
///////////////////////////////////////////////////////////////////////////
// Implementierung:
///////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <math.h>
#include <gmp.h>


namespace numbpart
{

typedef mpz_t* Pmpz_t;
int size;
Pmpz_t *p;

inline int omega(int m)
{
  return (3*m*m-m)/2;
} 

void numbpart_recurse(int n)
{
  if (n<0) { cerr << "invalid n in numbpart" << endl; return; }
  if (p[n]) return; // value is cached
  if (n==0) { p[n]=new mpz_t[1]; mpz_init_set_si(*p[n],1); return; }

  p[n]=new mpz_t[1];
  mpz_init(*p[n]);
  int m,i;
  m=1;
  while ((i=n-omega(m))>=0)
   {
     numbpart_recurse(i);
     if (m%2) mpz_add(*p[n],*p[n],*p[i]); else mpz_sub(*p[n],*p[n],*p[i]);
     m++;
   }
  m=1;
  while ((i=n-omega(-m))>=0)
   {
     numbpart_recurse(i);
     if (m%2) mpz_add(*p[n],*p[n],*p[i]); else mpz_sub(*p[n],*p[n],*p[i]);
     m++;
   }
}

void numbpart(mpz_t &res, int n)
{
  size=n+1;
  p = new Pmpz_t [size];
  for (int i=0; i<size; ++i) p[i]=NULL;
  numbpart_recurse(n);
  mpz_set(res,*p[n]);
  for (int i=0; i<size; ++i) if (p[i]) { mpz_clear(*p[i]); p[i]=NULL; }
  size=0;
  delete [] p;
}

}

#if 1
int main()
{
  mpz_t res;
  mpz_init(res);
  for (unsigned int i=10000; i<=11000; i+=101)
   {
     numbpart::numbpart(res,i);
     cout << "p(" << i << ") = ";
     mpz_out_str(stdout,10,res);
     cout << endl;
   }
  mpz_clear(res);
}
#endif
